@@ -24,6 +24,7 @@
 
 #include "soup-websocket-connection.h"
 #include "soup-enum-types.h"
+#include "soup-message.h"
 #include "soup-uri.h"
 
 /*
@@ -66,7 +67,6 @@
  * @error: default handler for the #SoupWebsocketConnection::error signal
  * @closing: the default handler for the #SoupWebsocketConnection:closing signal
  * @closed: default handler for the #SoupWebsocketConnection::closed signal
- * @pong: default handler for the #SoupWebsocketConnection::pong signal
  *
  * The abstract base class for #SoupWebsocketConnection
  *
@@ -81,8 +81,6 @@ enum {
 	PROP_ORIGIN,
 	PROP_PROTOCOL,
 	PROP_STATE,
-	PROP_MAX_INCOMING_PAYLOAD_SIZE,
-	PROP_KEEPALIVE_INTERVAL,
 };
 
 enum {
@@ -90,24 +88,16 @@ enum {
 	ERROR,
 	CLOSING,
 	CLOSED,
-	PONG,
 	NUM_SIGNALS
 };
 
 static guint signals[NUM_SIGNALS] = { 0, };
 
-typedef enum {
-	SOUP_WEBSOCKET_QUEUE_NORMAL = 0,
-	SOUP_WEBSOCKET_QUEUE_URGENT = 1 << 0,
-	SOUP_WEBSOCKET_QUEUE_LAST = 1 << 1,
-} SoupWebsocketQueueFlags;
-
 typedef struct {
 	GBytes *data;
+	gboolean last;
 	gsize sent;
 	gsize amount;
-	SoupWebsocketQueueFlags flags;
-	gboolean pending;
 } Frame;
 
 struct _SoupWebsocketConnectionPrivate {
@@ -116,8 +106,6 @@ struct _SoupWebsocketConnectionPrivate {
 	SoupURI *uri;
 	char *origin;
 	char *protocol;
-	guint64 max_incoming_payload_size;
-	guint keepalive_interval;
 
 	gushort peer_close_code;
 	char *peer_close_data;
@@ -142,19 +130,21 @@ struct _SoupWebsocketConnectionPrivate {
 	/* Current message being assembled */
 	guint8 message_opcode;
 	GByteArray *message_data;
-
-	GSource *keepalive_timeout;
 };
 
-#define MAX_INCOMING_PAYLOAD_SIZE_DEFAULT   128 * 1024
+#define MAX_PAYLOAD   128 * 1024
 
-G_DEFINE_TYPE_WITH_PRIVATE (SoupWebsocketConnection, soup_websocket_connection, G_TYPE_OBJECT)
+G_DEFINE_TYPE (SoupWebsocketConnection, soup_websocket_connection, G_TYPE_OBJECT)
+
+typedef enum {
+	SOUP_WEBSOCKET_QUEUE_NORMAL = 0,
+	SOUP_WEBSOCKET_QUEUE_URGENT = 1 << 0,
+	SOUP_WEBSOCKET_QUEUE_LAST = 1 << 1,
+} SoupWebsocketQueueFlags;
 
 static void queue_frame (SoupWebsocketConnection *self, SoupWebsocketQueueFlags flags,
 			 gpointer data, gsize len, gsize amount);
 
-static void protocol_error_and_close (SoupWebsocketConnection *self);
-
 static void
 frame_free (gpointer data)
 {
@@ -171,7 +161,8 @@ soup_websocket_connection_init (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv;
 
-	pv = self->pv = soup_websocket_connection_get_instance_private (self);
+	pv = self->pv = G_TYPE_INSTANCE_GET_PRIVATE (self, SOUP_TYPE_WEBSOCKET_CONNECTION,
+						     SoupWebsocketConnectionPrivate);
 
 	pv->incoming = g_byte_array_sized_new (1024);
 	g_queue_init (&pv->outgoing);
@@ -232,18 +223,6 @@ stop_output (SoupWebsocketConnection *self)
 	}
 }
 
-static void
-keepalive_stop_timeout (SoupWebsocketConnection *self)
-{
-	SoupWebsocketConnectionPrivate *pv = self->pv;
-
-	if (pv->keepalive_timeout) {
-		g_source_destroy (pv->keepalive_timeout);
-		g_source_unref (pv->keepalive_timeout);
-		pv->keepalive_timeout = NULL;
-	}
-}
-
 static void
 close_io_stop_timeout (SoupWebsocketConnection *self)
 {
@@ -261,7 +240,6 @@ close_io_stream (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 
-	keepalive_stop_timeout (self);
 	close_io_stop_timeout (self);
 
 	if (!pv->io_closing) {
@@ -352,21 +330,15 @@ send_message (SoupWebsocketConnection *self,
 	guint8 *mask = 0;
 	guint8 *at;
 
-	if (!(soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_OPEN)) {
-		g_debug ("Ignoring message since the connection is closed or is closing");
-		return;
-	}
-
 	bytes = g_byte_array_sized_new (14 + length);
 	outer = bytes->data;
 	outer[0] = 0x80 | opcode;
 
-	/* If control message, check payload size */
+	/* If control message, truncate payload */
 	if (opcode & 0x08) {
 		if (length > 125) {
-			g_warning ("WebSocket control message payload exceeds size limit");
-			protocol_error_and_close (self);
-			return;
+			g_warning ("Truncating WebSocket control message payload");
+			length = 125;
 		}
 
 		buffered_amount = 0;
@@ -382,14 +354,10 @@ send_message (SoupWebsocketConnection *self,
 		bytes->len = 4;
 	} else {
 		outer[1] = 127; /* mask | 64-bit-len */
-#if GLIB_SIZEOF_SIZE_T > 4
 		outer[2] = (length >> 56) & 0xFF;
 		outer[3] = (length >> 48) & 0xFF;
 		outer[4] = (length >> 40) & 0xFF;
 		outer[5] = (length >> 32) & 0xFF;
-#else
-		outer[2] = outer[3] = outer[4] = outer[5] = 0;
-#endif
 		outer[6] = (length >> 24) & 0xFF;
 		outer[7] = (length >> 16) & 0xFF;
 		outer[8] = (length >> 8) & 0xFF;
@@ -401,10 +369,9 @@ send_message (SoupWebsocketConnection *self,
 	 * probably a client somewhere that's not expecting it.
 	 */
 	if (self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_CLIENT) {
-		guint32 rnd = g_random_int ();
 		outer[1] |= 0x80;
 		mask = outer + bytes->len;
-		memcpy (mask, &rnd, sizeof (rnd));
+		* ((guint32 *)mask) = g_random_int ();
 		bytes->len += 4;
 	}
 
@@ -439,8 +406,6 @@ send_close (SoupWebsocketConnection *self,
 
 	send_message (self, flags, 0x08, (guint8 *)buffer, len);
 	self->pv->close_sent = TRUE;
-
-	keepalive_stop_timeout (self);
 }
 
 static void
@@ -523,7 +488,7 @@ bad_data_error_and_close (SoupWebsocketConnection *self)
 
 static void
 too_big_error_and_close (SoupWebsocketConnection *self,
-                         guint64 payload_len)
+                         gsize payload_len)
 {
 	GError *error;
 
@@ -532,72 +497,15 @@ too_big_error_and_close (SoupWebsocketConnection *self,
 				     self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER ?
 				     "Received extremely large WebSocket data from the client" :
 				     "Received extremely large WebSocket data from the server");
-	g_debug ("%s is trying to frame of size %" G_GUINT64_FORMAT " or greater, but max supported size is %" G_GUINT64_FORMAT,
+	g_debug ("%s is trying to frame of size %" G_GUINT64_FORMAT " or greater, but max supported size is 128KiB",
 		 self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER ? "server" : "client",
-	         payload_len, self->pv->max_incoming_payload_size);
+		 payload_len);
 	emit_error_and_close (self, error, TRUE);
 
 	/* The input is in an invalid state now */
 	stop_input (self);
 }
 
-static void
-close_connection (SoupWebsocketConnection *self,
-                  gushort                  code,
-                  const char              *data)
-{
-	SoupWebsocketQueueFlags flags;
-	SoupWebsocketConnectionPrivate *pv;
-
-	pv = self->pv;
-
-	if (pv->close_sent) {
-		g_debug ("close code already sent");
-		return;
-	}
-
-	/* Validate the closing code received by the peer */
-	switch (code) {
-	case SOUP_WEBSOCKET_CLOSE_NORMAL:
-	case SOUP_WEBSOCKET_CLOSE_GOING_AWAY:
-	case SOUP_WEBSOCKET_CLOSE_PROTOCOL_ERROR:
-	case SOUP_WEBSOCKET_CLOSE_UNSUPPORTED_DATA:
-	case SOUP_WEBSOCKET_CLOSE_BAD_DATA:
-	case SOUP_WEBSOCKET_CLOSE_POLICY_VIOLATION:
-	case SOUP_WEBSOCKET_CLOSE_TOO_BIG:
-		break;
-	case SOUP_WEBSOCKET_CLOSE_NO_EXTENSION:
-		if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER) {
-			g_debug ("Wrong closing code %d received for a server connection",
-			         code);
-		}
-		break;
-	case SOUP_WEBSOCKET_CLOSE_SERVER_ERROR:
-		if (pv->connection_type != SOUP_WEBSOCKET_CONNECTION_SERVER) {
-			g_debug ("Wrong closing code %d received for a non server connection",
-			         code);
-		}
-		break;
-	default:
-		if (code < 3000) {
-			g_debug ("Wrong closing code %d received", code);
-			protocol_error_and_close (self);
-			return;
-		}
-	}
-
-	g_signal_emit (self, signals[CLOSING], 0);
-
-	if (pv->close_received)
-		g_debug ("responding to close request");
-
-	flags = 0;
-	if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER && pv->close_received)
-		flags |= SOUP_WEBSOCKET_QUEUE_LAST;
-	send_close (self, flags, code, data);
-	close_io_after_timeout (self);
-}
-
 static void
 receive_close (SoupWebsocketConnection *self,
 	       const guint8 *data,
@@ -610,32 +518,17 @@ receive_close (SoupWebsocketConnection *self,
 	pv->peer_close_data = NULL;
 	pv->close_received = TRUE;
 
-	switch (len) {
-	case 0:
-		/* Send a clean close when having an empty payload */
-		close_connection (self, 1000, NULL);
-		return;
-	case 1:
-		/* Send a protocol error since the close code is incomplete */
-		protocol_error_and_close (self);
-		return;
-	default:
-		/* Store the code/data payload */
+	/* Store the code/data payload */
+	if (len >= 2) {
 		pv->peer_close_code = (guint16)data[0] << 8 | data[1];
-		break;
 	}
-
 	if (len > 2) {
 		data += 2;
 		len -= 2;
-		
-		if (!g_utf8_validate ((char *)data, len, NULL)) {
+		if (g_utf8_validate ((char *)data, len, NULL))
+			pv->peer_close_data = g_strndup ((char *)data, len);
+		else
 			g_debug ("received non-UTF8 close data: %d '%.*s' %d", (int)len, (int)len, (char *)data, (int)data[0]);
-			protocol_error_and_close (self);
-			return;
-		}
-
-		pv->peer_close_data = g_strndup ((char *)data, len);
 	}
 
 	/* Once we receive close response on server, close immediately */
@@ -644,7 +537,8 @@ receive_close (SoupWebsocketConnection *self,
 		if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER)
 			close_io_stream (self);
 	} else {
-		close_connection (self, pv->peer_close_code, pv->peer_close_data);
+		/* Send back the response */
+		soup_websocket_connection_close (self, pv->peer_close_code, NULL);
 	}
 }
 
@@ -658,27 +552,6 @@ receive_ping (SoupWebsocketConnection *self,
 	send_message (self, SOUP_WEBSOCKET_QUEUE_URGENT, 0x0A, data, len);
 }
 
-static void
-receive_pong (SoupWebsocketConnection *self,
-                      const guint8 *data,
-                      gsize len)
-{
-	GByteArray *bytes;
-
-	g_debug ("received pong message");
-
-	bytes = g_byte_array_sized_new (len + 1);
-	g_byte_array_append (bytes, data, len);
-	/* Always null terminate, as a convenience */
-	g_byte_array_append (bytes, (guchar *)"\0", 1);
-	/* But don't include the null terminator in the byte count */
-	bytes->len--;
-
-	g_signal_emit (self, signals[PONG], 0, bytes);
-	g_byte_array_unref (bytes);
-
-}
-
 static void
 process_contents (SoupWebsocketConnection *self,
 		  gboolean control,
@@ -708,12 +581,10 @@ process_contents (SoupWebsocketConnection *self,
 			receive_ping (self, payload, payload_len);
 			break;
 		case 0x0A:
-			receive_pong (self, payload, payload_len);
 			break;
 		default:
 			g_debug ("received unsupported control frame: %d", (int)opcode);
-			protocol_error_and_close (self);
-			return;
+			break;
 		}
 	} else if (pv->close_received) {
 		g_debug ("received message after close was received");
@@ -723,11 +594,11 @@ process_contents (SoupWebsocketConnection *self,
 		if (!fin && opcode) {
 			/* Initial fragment of a message */
 			if (pv->message_data) {
-				g_debug ("received out of order initial message fragment");
+				g_debug ("received out of order inital message fragment");
 				protocol_error_and_close (self);
 				return;
 			}
-			g_debug ("received initial fragment frame %d with %d payload", (int)opcode, (int)payload_len);
+			g_debug ("received inital fragment frame %d with %d payload", (int)opcode, (int)payload_len);
 		} else if (!fin && !opcode) {
 			/* Middle fragment of a message */
 			if (!pv->message_data) {
@@ -757,27 +628,12 @@ process_contents (SoupWebsocketConnection *self,
 
 		if (opcode) {
 			pv->message_opcode = opcode;
-			pv->message_data = g_byte_array_sized_new (payload_len + 1);
+			pv->message_data = g_byte_array_sized_new (payload_len);
 		}
 
 		switch (pv->message_opcode) {
 		case 0x01:
-		case 0x02:
-			g_byte_array_append (pv->message_data, payload, payload_len);
-			break;
-		default:
-			g_debug ("received unknown data frame: %d", (int)opcode);
-			protocol_error_and_close (self);
-			return;
-		}
-
-		/* Actually deliver the message? */
-		if (fin) {
-			if (pv->message_opcode == 0x01 &&
-			    !g_utf8_validate((char *)pv->message_data->data,
-			                     pv->message_data->len,
-			                     NULL)) {
-
+			if (!g_utf8_validate ((char *)payload, payload_len, NULL)) {
 				g_debug ("received invalid non-UTF8 text data");
 
 				/* Discard the entire message */
@@ -788,7 +644,17 @@ process_contents (SoupWebsocketConnection *self,
 				bad_data_error_and_close (self);
 				return;
 			}
+			/* fall through */
+		case 0x02:
+			g_byte_array_append (pv->message_data, payload, payload_len);
+			break;
+		default:
+			g_debug ("received unknown data frame: %d", (int)opcode);
+			break;
+		}
 
+		/* Actually deliver the message? */
+		if (fin) {
 			/* Always null terminate, as a convenience */
 			g_byte_array_append (pv->message_data, (guchar *)"\0", 1);
 
@@ -831,11 +697,6 @@ process_frame (SoupWebsocketConnection *self)
 	opcode = header[0] & 0x0f;
 	masked = ((header[1] & 0x80) != 0);
 
-	/* We do not support extensions, reserved bits must be 0 */
-	if (header[0] & 0x70) {
-		protocol_error_and_close (self);
-	}
-
 	switch (header[1] & 0x7f) {
 	case 126:
 		at = 4;
@@ -864,8 +725,7 @@ process_frame (SoupWebsocketConnection *self)
 	}
 
 	/* Safety valve */
-	if (self->pv->max_incoming_payload_size > 0 &&
-	    payload_len >= self->pv->max_incoming_payload_size) {
+	if (payload_len >= MAX_PAYLOAD) {
 		too_big_error_and_close (self, payload_len);
 		return FALSE;
 	}
@@ -965,12 +825,6 @@ on_web_socket_output (GObject *pollable_stream,
 	gssize count;
 	gsize len;
 
-	if (soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_CLOSED) {
-		g_debug ("Ignoring message since the connection is closed");
-		stop_output (self);
-		return TRUE;
-	}
-
 	frame = g_queue_peek_head (&pv->outgoing);
 
 	/* No more frames to send */
@@ -992,9 +846,6 @@ on_web_socket_output (GObject *pollable_stream,
 		if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
 			g_clear_error (&error);
 			count = 0;
-
-			g_debug ("failed to send frame because it would block, marking as pending");
-			frame->pending = TRUE;
 		} else {
 			emit_error_and_close (self, error, TRUE);
 			return FALSE;
@@ -1006,7 +857,7 @@ on_web_socket_output (GObject *pollable_stream,
 		g_debug ("sent frame");
 		g_queue_pop_head (&pv->outgoing);
 
-		if (frame->flags & SOUP_WEBSOCKET_QUEUE_LAST) {
+		if (frame->last) {
 			if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER) {
 				close_io_stream (self);
 			} else {
@@ -1043,6 +894,7 @@ queue_frame (SoupWebsocketConnection *self,
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	Frame *frame;
+	Frame *prev;
 
 	g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
 	g_return_if_fail (pv->close_sent == FALSE);
@@ -1052,22 +904,21 @@ queue_frame (SoupWebsocketConnection *self,
 	frame = g_slice_new0 (Frame);
 	frame->data = g_bytes_new_take (data, len);
 	frame->amount = amount;
-	frame->flags = flags;
+	frame->last = (flags & SOUP_WEBSOCKET_QUEUE_LAST) ? TRUE : FALSE;
 
 	/* If urgent put at front of queue */
 	if (flags & SOUP_WEBSOCKET_QUEUE_URGENT) {
-		GList *l;
-
-		/* Find out the first frame that is not urgent or partially sent or pending */
-		for (l = g_queue_peek_head_link (&pv->outgoing); l != NULL; l = l->next) {
-			Frame *prev = l->data;
-
-			if (!(prev->flags & SOUP_WEBSOCKET_QUEUE_URGENT) &&
-			    prev->sent == 0 && !prev->pending)
-				break;
+		/* But we can't interrupt a message already partially sent */
+		prev = g_queue_pop_head (&pv->outgoing);
+		if (prev == NULL) {
+			g_queue_push_head (&pv->outgoing, frame);
+		} else if (prev->sent > 0) {
+			g_queue_push_head (&pv->outgoing, frame);
+			g_queue_push_head (&pv->outgoing, prev);
+		} else {
+			g_queue_push_head (&pv->outgoing, prev);
+			g_queue_push_head (&pv->outgoing, frame);
 		}
-
-		g_queue_insert_before (&pv->outgoing, l, frame);
 	} else {
 		g_queue_push_tail (&pv->outgoing, frame);
 	}
@@ -1109,7 +960,6 @@ soup_websocket_connection_get_property (GObject *object,
 					GParamSpec *pspec)
 {
 	SoupWebsocketConnection *self = SOUP_WEBSOCKET_CONNECTION (object);
-	SoupWebsocketConnectionPrivate *pv = self->pv;
 
 	switch (prop_id) {
 	case PROP_IO_STREAM:
@@ -1136,14 +986,6 @@ soup_websocket_connection_get_property (GObject *object,
 		g_value_set_enum (value, soup_websocket_connection_get_state (self));
 		break;
 
-	case PROP_MAX_INCOMING_PAYLOAD_SIZE:
-		g_value_set_uint64 (value, pv->max_incoming_payload_size);
-		break;
-
-	case PROP_KEEPALIVE_INTERVAL:
-		g_value_set_uint (value, pv->keepalive_interval);
-		break;
-
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -1184,15 +1026,6 @@ soup_websocket_connection_set_property (GObject *object,
 		pv->protocol = g_value_dup_string (value);
 		break;
 
-	case PROP_MAX_INCOMING_PAYLOAD_SIZE:
-		pv->max_incoming_payload_size = g_value_get_uint64 (value);
-		break;
-
-	case PROP_KEEPALIVE_INTERVAL:
-		soup_websocket_connection_set_keepalive_interval (self,
-		                                                  g_value_get_uint (value));
-		break;
-
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -1231,7 +1064,6 @@ soup_websocket_connection_finalize (GObject *object)
 	g_assert (pv->io_closing);
 	g_assert (pv->io_closed);
 	g_assert (!pv->close_timeout);
-	g_assert (!pv->keepalive_timeout);
 
 	if (pv->message_data)
 		g_byte_array_free (pv->message_data, TRUE);
@@ -1249,6 +1081,8 @@ soup_websocket_connection_class_init (SoupWebsocketConnectionClass *klass)
 {
 	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
+	g_type_class_add_private (klass, sizeof (SoupWebsocketConnectionPrivate));
+
 	gobject_class->constructed = soup_websocket_connection_constructed;
 	gobject_class->get_property = soup_websocket_connection_get_property;
 	gobject_class->set_property = soup_websocket_connection_set_property;
@@ -1359,45 +1193,6 @@ soup_websocket_connection_class_init (SoupWebsocketConnectionClass *klass)
 							    G_PARAM_READABLE |
 							    G_PARAM_STATIC_STRINGS));
 
-	/**
-	 * SoupWebsocketConnection:max-incoming-payload-size:
-	 *
-	 * The maximum payload size for incoming packets the protocol expects
-	 * or 0 to not limit it.
-	 *
-	 * Since: 2.56
-	 */
-	g_object_class_install_property (gobject_class, PROP_MAX_INCOMING_PAYLOAD_SIZE,
-					 g_param_spec_uint64 ("max-incoming-payload-size",
-							      "Max incoming payload size",
-							      "Max incoming payload size ",
-							      0,
-							      G_MAXUINT64,
-							      MAX_INCOMING_PAYLOAD_SIZE_DEFAULT,
-							      G_PARAM_READWRITE |
-							      G_PARAM_CONSTRUCT |
-							      G_PARAM_STATIC_STRINGS));
-
-	/**
-	 * SoupWebsocketConnection:keepalive-interval:
-	 *
-	 * Interval in seconds on when to send a ping message which will
-	 * serve as a keepalive message. If set to 0 the keepalive message is
-	 * disabled.
-	 *
-	 * Since: 2.58
-	 */
-	g_object_class_install_property (gobject_class, PROP_KEEPALIVE_INTERVAL,
-					 g_param_spec_uint ("keepalive-interval",
-					                    "Keepalive interval",
-					                    "Keepalive interval",
-					                    0,
-					                    G_MAXUINT,
-					                    0,
-					                    G_PARAM_READWRITE |
-					                    G_PARAM_CONSTRUCT |
-					                    G_PARAM_STATIC_STRINGS));
-
 	/**
 	 * SoupWebsocketConnection::message:
 	 * @self: the WebSocket
@@ -1471,27 +1266,6 @@ soup_websocket_connection_class_init (SoupWebsocketConnectionClass *klass)
 					G_STRUCT_OFFSET (SoupWebsocketConnectionClass, closed),
 					NULL, NULL, g_cclosure_marshal_generic,
 					G_TYPE_NONE, 0);
-
-	/**
-	 * SoupWebsocketConnection::pong:
-	 * @self: the WebSocket
-	 * @message: the application data (if any)
-	 *
-	 * Emitted when we receive a Pong frame (solicited or
-	 * unsolicited) from the peer.
-	 *
-	 * As a convenience, the @message data will always be
-	 * NUL-terminated, but the NUL byte will not be included in
-	 * the length count.
-	 *
-	 * Since: 2.60
-	 */
-	signals[PONG] = g_signal_new ("pong",
-				      SOUP_TYPE_WEBSOCKET_CONNECTION,
-				      G_SIGNAL_RUN_FIRST,
-				      G_STRUCT_OFFSET (SoupWebsocketConnectionClass, pong),
-				      NULL, NULL, g_cclosure_marshal_generic,
-				      G_TYPE_NONE, 1, G_TYPE_BYTES);
 }
 
 /**
@@ -1766,6 +1540,7 @@ soup_websocket_connection_close (SoupWebsocketConnection *self,
 				 gushort code,
 				 const char *data)
 {
+	SoupWebsocketQueueFlags flags;
 	SoupWebsocketConnectionPrivate *pv;
 
 	g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
@@ -1780,119 +1555,14 @@ soup_websocket_connection_close (SoupWebsocketConnection *self,
 	else
 		g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_SERVER_ERROR);
 
-	close_connection (self, code, data);
-}
+	g_signal_emit (self, signals[CLOSING], 0);
 
-/**
- * soup_websocket_connection_get_max_incoming_payload_size:
- * @self: the WebSocket
- *
- * Gets the maximum payload size allowed for incoming packets.
- *
- * Returns: the maximum payload size.
- *
- * Since: 2.56
- */
-guint64
-soup_websocket_connection_get_max_incoming_payload_size (SoupWebsocketConnection *self)
-{
-	SoupWebsocketConnectionPrivate *pv;
+	if (pv->close_received)
+		g_debug ("responding to close request");
 
-	g_return_val_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self), MAX_INCOMING_PAYLOAD_SIZE_DEFAULT);
-	pv = self->pv;
-
-	return pv->max_incoming_payload_size;
-}
-
-/**
- * soup_websocket_connection_set_max_incoming_payload_size:
- * @self: the WebSocket
- * @max_incoming_payload_size: the maximum payload size
- *
- * Sets the maximum payload size allowed for incoming packets. It
- * does not limit the outgoing packet size.
- *
- * Since: 2.56
- */
-void
-soup_websocket_connection_set_max_incoming_payload_size (SoupWebsocketConnection *self,
-                                                         guint64                  max_incoming_payload_size)
-{
-	SoupWebsocketConnectionPrivate *pv;
-
-	g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
-	pv = self->pv;
-
-	if (pv->max_incoming_payload_size != max_incoming_payload_size) {
-		pv->max_incoming_payload_size = max_incoming_payload_size;
-		g_object_notify (G_OBJECT (self), "max-incoming-payload-size");
-	}
-}
-
-/**
- * soup_websocket_connection_get_keepalive_interval:
- * @self: the WebSocket
- *
- * Gets the keepalive interval in seconds or 0 if disabled.
- *
- * Returns: the keepalive interval.
- *
- * Since: 2.58
- */
-guint
-soup_websocket_connection_get_keepalive_interval (SoupWebsocketConnection *self)
-{
-	SoupWebsocketConnectionPrivate *pv;
-
-	g_return_val_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self), 0);
-	pv = self->pv;
-
-	return pv->keepalive_interval;
-}
-
-static gboolean
-on_queue_ping (gpointer user_data)
-{
-	SoupWebsocketConnection *self = SOUP_WEBSOCKET_CONNECTION (user_data);
-	static const char ping_payload[] = "libsoup";
-
-	g_debug ("sending ping message");
-
-	send_message (self, SOUP_WEBSOCKET_QUEUE_NORMAL, 0x09,
-		      (guint8 *) ping_payload, strlen(ping_payload));
-
-	return G_SOURCE_CONTINUE;
-}
-
-/**
- * soup_websocket_connection_set_keepalive_interval:
- * @self: the WebSocket
- * @interval: the interval to send a ping message or 0 to disable it
- *
- * Sets the interval in seconds on when to send a ping message which will serve
- * as a keepalive message. If set to 0 the keepalive message is disabled.
- *
- * Since: 2.58
- */
-void
-soup_websocket_connection_set_keepalive_interval (SoupWebsocketConnection *self,
-                                                  guint                    interval)
-{
-	SoupWebsocketConnectionPrivate *pv;
-
-	g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
-	pv = self->pv;
-
-	if (pv->keepalive_interval != interval) {
-		pv->keepalive_interval = interval;
-		g_object_notify (G_OBJECT (self), "keepalive-interval");
-
-		keepalive_stop_timeout (self);
-
-		if (interval > 0) {
-			pv->keepalive_timeout = g_timeout_source_new_seconds (interval);
-			g_source_set_callback (pv->keepalive_timeout, on_queue_ping, self, NULL);
-			g_source_attach (pv->keepalive_timeout, pv->main_context);
-		}
-	}
+	flags = 0;
+	if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER && pv->close_received)
+		flags |= SOUP_WEBSOCKET_QUEUE_LAST;
+	send_close (self, flags, code, data);
+	close_io_after_timeout (self);
 }
